/*
 * ====================================================================
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 */

package org.apache.hc.core5.concurrent;

import java.util.Objects;
import java.util.function.Function;

/**
 * Completes a result on a {@link BasicFuture}.
 *
 * @param <T> the future result type of an asynchronous operation for this type.
 * @param <U> the future result type of an asynchronous operation for the {@link FutureCallback}.
 * @since 5.4
 */
public class CompletingFutureContribution<T, U> extends FutureContribution<T> {

    private final Function<T, U> resultProvider;

    /**
     * Constructs a new instance to callback the given {@link BasicFuture}.
     *
     * @param future The callback.
     */
    @SuppressWarnings("unchecked")
    public CompletingFutureContribution(final BasicFuture<U> future) {
        this(future, (Function<T, U>) Function.identity());
    }

    /**
     * Constructs a new instance to callback the given {@link BasicFuture}.
     *
     * @param future The callback.
     * @param resultProvider Provides the result to complete the future.
     */
    public CompletingFutureContribution(final BasicFuture<U> future, final Function<T, U> resultProvider) {
        super(future);
        this.resultProvider = Objects.requireNonNull(resultProvider, "resultProvider");
    }

    @Override
    public void completed(final T result) {
        getFuture().completed(resultProvider.apply(result));
    }

}
