/*
 * ====================================================================
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 */
package org.apache.hc.core5.reactor;

import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.Mockito;

class IOWorkerSelectorsTest {

    @ParameterizedTest(name = "worker count = {0}")
    @ValueSource(ints = {1, 2, 3, 4, 5, 10, 15, 16, 32})
    void testIndexOverflow(final int workerCount) {
        final long start = (long) Integer.MAX_VALUE - 10;
        final long end = (long) Integer.MAX_VALUE + 10;
        final IOWorkerStats[] workers = new IOWorkerStats[workerCount];
        for (int i = 0; i < workerCount; i++) {
            workers[i] = Mockito.mock(IOWorkerStats.class);
        }
        final IOWorkerSelector selector = IOWorkerSelectors.newSelector(workerCount, (int) start);
        for (long i = start; i < end; i++) {
            Assertions.assertTrue(selector.select(workers) < workerCount);
        }
    }

}
