/*
 * ====================================================================
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 */

package org.apache.hc.core5.testing.nio;

import org.apache.hc.core5.http2.ssl.ConscryptClientTlsStrategy;
import org.apache.hc.core5.http2.ssl.ConscryptServerTlsStrategy;
import org.apache.hc.core5.http2.ssl.H2ClientTlsStrategy;
import org.apache.hc.core5.http2.ssl.H2ServerTlsStrategy;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.condition.DisabledOnOs;
import org.junit.jupiter.api.condition.EnabledForJreRange;
import org.junit.jupiter.api.condition.JRE;
import org.junit.jupiter.api.condition.OS;

class AlpnTests {

    @Nested
    @DisplayName("ALPN Oracle")
    class OracleJSSEAlpnTest extends AlpnTest {

        public OracleJSSEAlpnTest() throws Exception {
            super("Oracle", H2ServerTlsStrategy::new, H2ClientTlsStrategy::new);
        }

    }

    @Nested
    @DisplayName("ALPN Conscrypt (Java 9 or newer)")
    @DisabledOnOs(OS.MAC)
    @EnabledForJreRange(min = JRE.JAVA_9)
    class ConscryptJSSEAlpnTest extends AlpnTest {

        public ConscryptJSSEAlpnTest() throws Exception {
            super("Conscrypt", H2ServerTlsStrategy::new, H2ClientTlsStrategy::new);
        }

    }

    @Nested
    @DisplayName("ALPN Conscrypt (Conscrypt specific TLS strategies)")
    @DisabledOnOs(OS.MAC)
    class ConscryptJSSEAndStrategiesAlpnTest extends AlpnTest {

        public ConscryptJSSEAndStrategiesAlpnTest() throws Exception {
            super("Conscrypt", ConscryptServerTlsStrategy::new, ConscryptClientTlsStrategy::new);
        }

    }

}
